//
//  TPGCodec.m
//  TPGTest
//
//  Created by narvyliu
//  Copyright (c) 2017年 narvyliu. All rights reserved.
//

#include <mach/mach_time.h>
#import "TPGDecTest.h"
#include "TPGDecoder.h"
#include <sys/stat.h>
#import <UIKit/UIKit.h>
#import "UIImage+TPG.h"

#include "imageUtil.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

@implementation TPGDecTest

#define NAL_TYPE_EXIF_INFO	0x1000

typedef struct
{
    //uint16_t    bfType;
    uint32_t    bfSize;
    uint16_t    bfReserved1;
    uint16_t    bfReserved2;
    uint32_t    bfOffBits;
} ClBitMapFileHeader;

typedef struct
{
    uint32_t   biSize;
    int32_t    biWidth;
    int32_t    biHeight;
    uint16_t   biPlanes;
    uint16_t   biBitCount;
    uint32_t   biCompression;
    uint32_t   biSizeImage;
    int32_t    biXPelsPerMeter;
    int32_t    biYPelsPerMeter;
    uint32_t   biClrUsed;
    uint32_t   biClrImportant;
} ClBitMapInfoHeader;

void write_bmp_header(FILE* fp, int width, int height)
{
    ClBitMapFileHeader bmpFileHeader;
    ClBitMapInfoHeader bmpInfoHeader;
    
    int image_width = width;
    int image_height = height;
    uint16_t fileType;
    fileType = 0x4D42;
    fwrite(&fileType, sizeof(uint16_t), 1, fp);
    
    bmpFileHeader.bfSize = image_width*image_height + 54;
    bmpFileHeader.bfReserved1 = 0;
    bmpFileHeader.bfReserved2 = 0;
    bmpFileHeader.bfOffBits = 54;
    fwrite(&bmpFileHeader, sizeof(ClBitMapFileHeader), 1, fp);
    
    bmpInfoHeader.biSize = 40;
    bmpInfoHeader.biWidth = image_width;
    bmpInfoHeader.biHeight = image_height;
    bmpInfoHeader.biPlanes = 1;
    bmpInfoHeader.biBitCount = 24;
    bmpInfoHeader.biCompression = 0;
    bmpInfoHeader.biSizeImage =image_height*image_width;
    bmpInfoHeader.biXPelsPerMeter = 0;
    bmpInfoHeader.biYPelsPerMeter = 0;
    bmpInfoHeader.biClrUsed = 0;
    bmpInfoHeader.biClrImportant = 0;
    fwrite(&bmpInfoHeader, sizeof(ClBitMapInfoHeader), 1, fp);
}



void write_bmp_data(FILE* fp, int width, int height, int components, unsigned char *src_buff)
{
    unsigned char *dst_width_buff;
    unsigned char *point;
    
    uint16_t depth;
    
    unsigned char temp;
    depth=components;
    
    dst_width_buff=new unsigned char[width*depth];
    memset(dst_width_buff,0,sizeof(unsigned char)*width*depth);
    
    point=src_buff+width*depth*(height-1);
    for (uint32_t i=0;i<height;i++)
    {
        for (uint32_t j=0;j<width*depth;j+=depth)
        {
            if (depth==1)
            {
                dst_width_buff[j]=point[j];
            }
            
            if (depth==3)
            {
                dst_width_buff[j+2]=point[j+0];
                dst_width_buff[j+1]=point[j+1];
                dst_width_buff[j+0]=point[j+2];
            }
        }
        point-=width*depth;
        fwrite(dst_width_buff,sizeof(unsigned char)*width*depth,1,fp);    //“ª¥Œ–¥“ª––
    }
    
    if (dst_width_buff)
    {
        delete[] dst_width_buff;
    }
}



static void saveBmp(FILE* fp, int width, int height, unsigned char* pData)
{
    write_bmp_header(fp, width, height);
    write_bmp_data(fp, width, height, 3, pData);
}


UIImage* decodeTPG2JPG(char * InputFile, char* OutputFile)
{
    FILE* fp = NULL;
    if ((fp = fopen(InputFile, "rb")) == NULL)
    {
        printf("can't open input file :\n");
        return nullptr;
    }
    fseek(fp, 0, SEEK_END);
    int length = ftell(fp);
    fseek(fp, 0, SEEK_SET);
    unsigned char* pStreamBuf = (unsigned char*)malloc(length);
    fread(pStreamBuf, length, 1, fp);
    fclose(fp);
    
    TPGFeatures features = {0};
    TPGStatusCode stats = TPGParseHeader(pStreamBuf, length, &features);
    if(stats != TPG_STATUS_OK)
    {
        return nullptr;
    }
    
    void* pDec = TPGDecCreate(pStreamBuf,length);
    
    const unsigned char* pInfo = NULL;
    int nlen = 0;
    stats = TPGGetAdditionalInfo(pDec,  pStreamBuf, length, NAL_TYPE_EXIF_INFO, &pInfo, &nlen);
    
    int width = features.width;
    int height = features.height;
    
    enRawDataFormat imgFormat = FORMAT_RGB;
    int bufsize = width*height*3;
    unsigned char* pOutData = (unsigned char*)malloc(bufsize);
    
    TPGOutFrame outFrame = {0};
    outFrame.dstWidth = width;
    outFrame.dstHeight = height;
    outFrame.pOutBuf = pOutData;
    outFrame.bufsize = bufsize;
    outFrame.fmt = FORMAT_RGB;
    
    mach_timebase_info_data_t info;
    volatile uint64_t starttime_c,sumtime_c=0;
    if (mach_timebase_info(&info) != KERN_SUCCESS)
    {
        printf("time unsuccess!\n");
    }
    
    
    
    starttime_c = mach_absolute_time();

    TPGDecodeImage(pDec, pStreamBuf, length, 0, &outFrame);
        
    sumtime_c += mach_absolute_time() - starttime_c;
    printf("%s, time: %lf ms\n",InputFile, (double)(sumtime_c*info.numer / info.denom)/NSEC_PER_MSEC);
        
    BitmapData bm = {0};
    bm.data = pOutData;
    bm.width = width;
    bm.height = height;
    bm.bytePerPix = 3;
    bm.componentsPerPix = 3;
    bm.bytePerRow = width*3;
    bm.PixFormat = JPEGCS_RGB;
    
    //encodeBitmapToFile(&bm, out_path, 80, 0);
    encodeBitmapToFileWithExif(&bm, OutputFile, 80, 0, pInfo, nlen);
    
    
    
    unsigned char* pOutDataRGBA = (unsigned char*)malloc(width*height*4);


    int i = 0;
    int j = 0;

    for(; i<width*height*4; i=i+4, j=j+3)
    {
        pOutDataRGBA[i] = pOutData[j];
        pOutDataRGBA[i+1] = pOutData[j+1];
        pOutDataRGBA[i+2] = pOutData[j+2];
        pOutDataRGBA[i+3] = 255;
    }

    UIImage* result = [UIImage convertBitmapRGBA8ToUIImage:pOutDataRGBA withWidth:width withHeight:height];
    
    free(pOutDataRGBA);
    pOutDataRGBA = NULL;
    
    /*
     BitmapData bitmap = {0};
     
     bitmap.width       = width;
     bitmap.height      = height;
     bitmap.bytePerPix  = 24;
     bitmap.componentsPerPix = 3;
     bitmap.bytePerRow = width * 3;
     bitmap.orientation = 0;
     bitmap.data        = (unsigned char*)malloc(width*height*3);
     bitmap.PixFormat   = JCS_YCbCr;
     
     unsigned char* ypixel, *upixel, *vpixel;
     
     unsigned int encWidth = (width+1)/2*2;
     unsigned int encHeight = (height+1)/2*2;
     
     ypixel = pOutData;
     upixel = ypixel + encWidth * encHeight;
     vpixel = upixel + encWidth * encHeight/4;
     
     unsigned char * outYUV  = bitmap.data ;
     
     for (int i = 0; i < height ; i++)
     {
     for (int j = 0; j < width ; j++)
     {
     int chromaX = j/2;
     int chromaY = i/2;
     *(outYUV + i*width*3 +j*3) = *(ypixel+ i*encWidth +j);
     *(outYUV + i*width*3 +j*3 + 1) = *(upixel+ chromaY*(encWidth/2) + chromaX);
     *(outYUV + i*width*3 +j*3 + 2) = *(vpixel+ chromaY*(encWidth/2) + chromaX);
     }
     }
     
     encodeBitmapToFile(&bitmap, OutputFile, 100, true);
     free(bitmap.data);
     */
    
    
    //float decodeTime = [[TimeMonitor instance] stopMonitor];
    
    TPGDecDestroy(pDec);
    pDec = NULL;
    
    free(pStreamBuf);
    pStreamBuf = NULL;
    free(pOutData);
    pOutData = NULL;
    return result;
}

UIImage* decodeTPG2PNG(char * InputFile, char* OutputFile)
{
    FILE* fp = NULL;
    if ((fp = fopen(InputFile, "rb")) == NULL)
    {
        printf("can't open input file :\n");
        return nullptr;
    }
    fseek(fp, 0, SEEK_END);
    int length = ftell(fp);
    fseek(fp, 0, SEEK_SET);
    unsigned char* pStreamBuf = (unsigned char*)malloc(length);
    fread(pStreamBuf, length, 1, fp);
    fclose(fp);
    
    TPGFeatures features = {0};
    TPGStatusCode stats = TPGParseHeader(pStreamBuf, length, &features);
    if(stats != TPG_STATUS_OK)
    {
        return nullptr;
    }
    
    void* pDec = TPGDecCreate(pStreamBuf,length);
    
    
    int width = features.width;
    int height = features.height;
    
    enRawDataFormat imgFormat = FORMAT_RGBA;
    int bufsize = width*height*4;
    unsigned char* pOutData = (unsigned char*)malloc(bufsize);
    
    
    TPGOutFrame outFrame = {0};
    outFrame.dstWidth = width;
    outFrame.dstHeight = height;
    outFrame.pOutBuf = pOutData;
    outFrame.bufsize = bufsize;
    outFrame.fmt = imgFormat;
    
    mach_timebase_info_data_t info;
    volatile uint64_t starttime_c,sumtime_c=0;
    if (mach_timebase_info(&info) != KERN_SUCCESS)
    {
        printf("time unsuccess!\n");
    }
    
    starttime_c = mach_absolute_time();
    
    TPGDecodeImage(pDec, pStreamBuf, length, 0, &outFrame);
    
    sumtime_c = mach_absolute_time() - starttime_c;
    printf("%s, time: %lf ms\n",InputFile, (double)(sumtime_c*info.numer / info.denom)/NSEC_PER_MSEC);
    
    
    pic_data outData;
    outData.width = width;
    outData.height = height;
    outData.rgba = pOutData;
    outData.flag = 1;
    outData.bit_depth = 8;
    
    if(encode_png(OutputFile, &outData))
    {
        printf("lodepng_decode32_file error!");
    }
    UIImage* result = [UIImage convertBitmapRGBA8ToUIImage:pOutData withWidth:width withHeight:height];
    /*
     BitmapData bitmap = {0};
     
     bitmap.width       = width;
     bitmap.height      = height;
     bitmap.bytePerPix  = 24;
     bitmap.componentsPerPix = 3;
     bitmap.bytePerRow = width * 3;
     bitmap.orientation = 0;
     bitmap.data        = (unsigned char*)malloc(width*height*3);
     bitmap.PixFormat   = JCS_YCbCr;
     
     unsigned char* ypixel, *upixel, *vpixel;
     
     unsigned int encWidth = (width+1)/2*2;
     unsigned int encHeight = (height+1)/2*2;
     
     ypixel = pOutData;
     upixel = ypixel + encWidth * encHeight;
     vpixel = upixel + encWidth * encHeight/4;
     
     unsigned char * outYUV  = bitmap.data ;
     
     for (int i = 0; i < height ; i++)
     {
     for (int j = 0; j < width ; j++)
     {
     int chromaX = j/2;
     int chromaY = i/2;
     *(outYUV + i*width*3 +j*3) = *(ypixel+ i*encWidth +j);
     *(outYUV + i*width*3 +j*3 + 1) = *(upixel+ chromaY*(encWidth/2) + chromaX);
     *(outYUV + i*width*3 +j*3 + 2) = *(vpixel+ chromaY*(encWidth/2) + chromaX);
     }
     }
     
     encodeBitmapToFile(&bitmap, OutputFile, 100, true);
     free(bitmap.data);
     */
    
    /*
     CGDataProviderRef provider = CGDataProviderCreateWithData(NULL, pOutData, bufsize, NULL);
     CGColorSpaceRef colorSpaceRef = CGColorSpaceCreateDeviceRGB();
     CGImageRef imageRef = CGImageCreate(width, height, 8, 24,
     3 * width, colorSpaceRef, kCGBitmapByteOrderDefault,
     provider, NULL, YES, kCGRenderingIntentDefault);
     
     CGColorSpaceRelease(colorSpaceRef);
     CGDataProviderRelease(provider);
     
     UIImage *newImage = [[UIImage alloc] initWithCGImage:imageRef];
     CGImageRelease(imageRef);
     UIImageJPEGRepresentation(newImage, 1.0);
     [UIImageJPEGRepresentation(newImage, 1.0) writeToFile:[[NSString alloc] initWithCString:(const char*)OutputFile encoding:NSASCIIStringEncoding] atomically:YES];
     //float decodeTime = [[TimeMonitor instance] stopMonitor];
     */
    
    TPGDecDestroy(pDec);
    pDec = NULL;

    free(pStreamBuf);
    pStreamBuf = NULL;
    free(pOutData);
    pOutData = NULL;
    return result;
    
}


UIImage* decodeTPGGIF(char *InputFile, char* OutputFile)
{
    unsigned char* pOutBuf, *pOutBuf32 = NULL;
    int nOutWidth;
    int nOutHeight;
    int nPixelSize;
    
    //decode TPG
    UIImage* result = nil;
    
    int stream_len = 0;
    FILE* fp = fopen(InputFile, "rb");
    struct stat fileInfo;
    if(stat(InputFile,&fileInfo))
    {
        printf("can't find the %s file.\n",InputFile);
        exit(1);
    }
    stream_len = fileInfo.st_size;
    unsigned char* pStreamBuf = (unsigned char*)malloc(stream_len);
    if(fread(pStreamBuf,1,stream_len,fp) !=stream_len)
    {
        printf("read the %s fail ");
        exit(1);
    }
    fclose(fp);
    
    enRawDataFormat fmt = enRawDataFormat::FORMAT_RGB;
    nPixelSize = 3;
    
    TPGFeatures features = {0};
    TPGStatusCode stats = TPGParseHeader(pStreamBuf, stream_len, &features);
    
    if (TPG_STATUS_OK != stats)
    {
        printf("parse TPG header info error!\n");
        exit(1);
    }
    
    if(features.image_mode == emMode_AnimationWithAlpha)
    {
        fmt = enRawDataFormat::FORMAT_RGBA;
        nPixelSize = 4;
    }
    
    
    void *TPGDec = TPGDecCreate(pStreamBuf, stream_len);
    
    nOutWidth = features.width;
    nOutHeight = features.height;
    pOutBuf = new unsigned char[(nOutWidth + 1) * (nOutHeight + 1) * nPixelSize];
    
    if(features.image_mode == emMode_Animation)
    {
        pOutBuf32 = new unsigned char[(nOutWidth + 1) * (nOutHeight + 1) * 4];
    }
    
    TPGOutFrame outFrame = {0};
    outFrame.dstWidth = nOutWidth;
    outFrame.dstHeight = nOutHeight;
    outFrame.pOutBuf = pOutBuf;
    outFrame.bufsize = (nOutWidth + 1) * (nOutHeight + 1) * nPixelSize;
    outFrame.fmt = fmt;
    
    mach_timebase_info_data_t info;
    volatile uint64_t starttime_c,sumtime_c=0;
    if (mach_timebase_info(&info) != KERN_SUCCESS)
    {
        printf("time unsuccess!\n");
    }
    
    NSMutableArray<UIImage *> *AnimatedImages = @[].mutableCopy;
    NSTimeInterval duration = 0;
    
    if (features.image_mode == emMode_Animation || features.image_mode == emMode_AnimationWithAlpha)
    {
        void* hGif = GifEncoderOpen(OutputFile, nOutWidth, nOutHeight, features.image_mode==emMode_AnimationWithAlpha, NULL,0);
        
        //FILE *fRGB = fopen(OutputFile,"wb");
        int i = 0;
        for (i = 0; i < features.frame_count; i++)
        {
            //starttime_c = mach_absolute_time();
            stats = TPGDecodeImage(TPGDec, pStreamBuf, stream_len, i, &outFrame);
            int time = -1;
            stats = TPGGetDelayTime(TPGDec, pStreamBuf, stream_len, i, &time);
            printf("time = %d\n", time);
            if (TPG_STATUS_OK != stats)
            {
                break;
            }
            
            UIImage* oneFrame = nil;
            
            if(features.image_mode == emMode_Animation)
            {
                int i = 0;
                int j = 0;
                
                for(; i<nOutWidth*nOutHeight*4; i=i+4, j=j+3)
                {
                    pOutBuf32[i] = pOutBuf[j];
                    pOutBuf32[i+1] = pOutBuf[j+1];
                    pOutBuf32[i+2] = pOutBuf[j+2];
                    pOutBuf32[i+3] = 255;
                }
                oneFrame = [UIImage convertBitmapRGBA8ToUIImage:pOutBuf32 withWidth:nOutWidth withHeight:nOutHeight];
            }
            else{
                oneFrame = [UIImage convertBitmapRGBA8ToUIImage:pOutBuf withWidth:nOutWidth withHeight:nOutHeight];
            }

            if(oneFrame)
            {
                [AnimatedImages addObject:oneFrame];
                if(outFrame.delayTime >=2)
                {
                    duration += outFrame.delayTime *10;
                }
                else{
                
                    duration +=100;
                }
            }
            
            starttime_c = mach_absolute_time();
            if (hGif)
            {
                stGifFrame stFrame = {0};
                stFrame.width = outFrame.dstWidth;
                stFrame.height = outFrame.dstHeight;
                stFrame.pRGB = outFrame.pOutBuf;
                stFrame.delayTime = outFrame.delayTime/10;
                GifEncoderEncodeFrame(hGif, &stFrame);
            }
            sumtime_c += mach_absolute_time() - starttime_c;

            //fwrite(outFrame.pOutBuf, outFrame.dstWidth*outFrame.dstHeight*nPixelSize, 1, fRGB);
        }
        
        printf("decode time: %s, %lf ms\n", InputFile, (double)(sumtime_c*info.numer / info.denom)/NSEC_PER_MSEC);
       // fclose(fRGB);
       GifEncoderClose(hGif);
    }
    
    if(AnimatedImages.count > 0)
    {
        result = [UIImage animatedImageWithImages: [AnimatedImages copy] duration:(duration/1000)];
    }
    
    
    free(pStreamBuf);
    pStreamBuf = NULL;
    delete [] pOutBuf;
    pOutBuf = NULL;
    if(pOutBuf32)
    {
        delete [] pOutBuf32;
        pOutBuf32 = NULL;
    }
    
    TPGDecDestroy(TPGDec);
    TPGDec = NULL;
    
    return result;
}

- (UIImage*) onTPGDecode
{

    UIImage *img;
    
    NSFileManager *manager = [NSFileManager defaultManager];
    NSString *NspathDocuments = [NSSearchPathForDirectoriesInDomains(NSDocumentDirectory, NSUserDomainMask, YES) objectAtIndex:0];
    NSDirectoryEnumerator *direnum = [manager enumeratorAtPath:NspathDocuments];
    
    NSMutableArray *files = [NSMutableArray arrayWithCapacity:42];
    NSString *filename ;
    while (filename = [direnum nextObject]) {
        if ([[filename pathExtension] isEqual:@"tpg"]) {
            [files addObject: filename];
        }
    }
    
    NSEnumerator *fileenum;
    fileenum = [files objectEnumerator];
    
    while (filename = [fileenum nextObject]) {
        
        //        NSLog(@"%@", filename);
        NSString *NsInputFileNamepath = [NSString stringWithFormat:@"%@/%@",NspathDocuments,filename];
        int len = [NsInputFileNamepath length];
        char * InputFile = (char*)malloc(sizeof(char)*len+1);
        [NsInputFileNamepath getCString:InputFile maxLength:len+1 encoding:NSUTF8StringEncoding];
        
        NSString *NsOutputFileNamepath = [NSString stringWithFormat:@"%@/%@",NspathDocuments,filename];
        len = [NsOutputFileNamepath length];
        char * OutputFile =(char*)malloc(sizeof(char)*len+1+4);
        [NsOutputFileNamepath getCString:OutputFile maxLength:len+1 encoding:NSUTF8StringEncoding];
        
        //decode TPG
        char name[4];
        
        int stream_len = 0;
        FILE* fp = fopen(InputFile, "rb");
        struct stat fileInfo;
        if(stat(InputFile,&fileInfo))
        {
            printf("can't find the %s file.\n",InputFile);
            return nullptr;
        }
        stream_len = fileInfo.st_size;
        unsigned char* pStreamBuf = (unsigned char*)malloc(stream_len);
        if(fread(pStreamBuf,1,stream_len,fp) !=stream_len)
        {
            printf("read the file %s failed\n", InputFile);
            return nullptr;
        }
        fclose(fp);
        
        TPGFeatures features = {0};
        TPGStatusCode stats = TPGParseHeader(pStreamBuf, stream_len, &features);
        
        if (TPG_STATUS_OK != stats)
        {
            printf("parse TPG header info error!\n");
            return nullptr;
        }
        
        if (features.image_mode == emMode_Normal || features.image_mode == emMode_BlendAlpha)
        {
            strcat(OutputFile, ".jpg");
            img = decodeTPG2JPG(InputFile, OutputFile);
        }
        else if (features.image_mode == emMode_EncodeAlpha)
        {
            strcat(OutputFile, ".png");
            img = decodeTPG2PNG(InputFile, OutputFile);
        }
        else if (features.image_mode == emMode_Animation || features.image_mode == emMode_AnimationWithAlpha)
        {
            strcat(OutputFile, ".gif");
            img = decodeTPGGIF(InputFile, OutputFile);
        }
        else
        {
            printf("TPG file format error!\n");
            return nullptr;
        }
      
        free(InputFile);
        InputFile = NULL;
        free(OutputFile);
        OutputFile = NULL;
        free(pStreamBuf);
        pStreamBuf = NULL;
    }
    return img;
}


- (UIImage*) onTPGDecode2: (char *)InputFile
               saveoutputFIle: (char *)OutputFile
{
    
        UIImage *img;

        
        //decode TPG
        char name[4];
        
        int stream_len = 0;
        FILE* fp = fopen(InputFile, "rb");
        struct stat fileInfo;
        if(stat(InputFile,&fileInfo))
        {
            printf("can't find the %s file.\n",InputFile);
            return nullptr;
        }
        stream_len = fileInfo.st_size;
        unsigned char* pStreamBuf = (unsigned char*)malloc(stream_len);
        if(fread(pStreamBuf,1,stream_len,fp) !=stream_len)
        {
            printf("read the file %s failed\n", InputFile);
            return nullptr;
        }
        fclose(fp);
        
        TPGFeatures features = {0};
        TPGStatusCode stats = TPGParseHeader(pStreamBuf, stream_len, &features);
        
        if (TPG_STATUS_OK != stats)
        {
            printf("parse TPG header info error!\n");
            return nullptr;
        }
        
        if (features.image_mode == emMode_Normal || features.image_mode == emMode_BlendAlpha)
        {
            strcat(OutputFile, ".jpg");
            img = decodeTPG2JPG(InputFile, OutputFile);
        }
        else if (features.image_mode == emMode_EncodeAlpha)
        {
            strcat(OutputFile, ".png");
            img = decodeTPG2PNG(InputFile, OutputFile);
        }
        else if (features.image_mode == emMode_Animation || features.image_mode == emMode_AnimationWithAlpha)
        {
            strcat(OutputFile, ".gif");
            img = decodeTPGGIF(InputFile, OutputFile);
        }
        else
        {
            printf("TPG file format error!\n");
            return nullptr;
        }
        
        free(InputFile);
        InputFile = NULL;
        free(OutputFile);
        OutputFile = NULL;
        free(pStreamBuf);
        pStreamBuf = NULL;
    
        return img;
}
@end
