#import<objc/runtime.h>
#import "UIImage+TPG.h"
#import<UIKit/UIKit.h>

#include "TPGDecoder.h"
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

@implementation UIImage(TPG)

+ (void)load{
    Method m1 = class_getClassMethod([UIImage class], @selector(imageWithContentsOfFile:));
    Method m2 = class_getClassMethod([UIImage class], @selector(TPGImageWithContentsOfFile:));
    method_exchangeImplementations(m1, m2);
}

+(UIImage*)TPGImageWithContentsOfFile:(NSString*)path{
    if ([UIImage isTPGFile:path]){
        int len = (int)[path length];
        char * InputFile = (char*)malloc(sizeof(char)*len+1);
        [path getCString:InputFile maxLength:len+1 encoding:NSUTF8StringEncoding];
        UIImage *result = [UIImage decodeTPG:InputFile];
        free(InputFile);
        InputFile = NULL;
        return result;
    }else{
        return [UIImage TPGImageWithContentsOfFile:path];
        NSData *imageData = [NSData dataWithContentsOfFile:path];
        return [UIImage imageWithData:imageData];
    }
}

+ (BOOL)isTPGFile:(NSString*)path{
    BOOL result = NO;
    BOOL isDir = NO;
    if (!path || ![[NSFileManager defaultManager] fileExistsAtPath:path isDirectory:&isDir]){
        return result;
    }
    int len = (int)[path length];
    char * InputFile = (char*)malloc(sizeof(char)*len+1);
    [path getCString:InputFile maxLength:len+1 encoding:NSUTF8StringEncoding];
    FILE* fp = fopen(InputFile, "rb");
    free(InputFile);
    InputFile = NULL;
    if (fp == NULL){
        return result;
    }
    
    NSString *TPGHead = @"TPG";
    char * cTPGHead = (char*)malloc(sizeof(char)*6+1);
    [TPGHead getCString:cTPGHead maxLength:6+1 encoding:NSUTF8StringEncoding];
    
    char* pStreamBuf = (char*)malloc(6+1);
    if(fread(pStreamBuf,1,6,fp) == 6){
        if (strcmp(pStreamBuf,cTPGHead) == 0){
            result = YES;
        }
    };
    fclose(fp);
    free(pStreamBuf);
    pStreamBuf = NULL;
    free(cTPGHead);
    cTPGHead = NULL;
    return result;
}

+(UIImage*)decodeTPG:(char *)InputFile{
    UIImage *result = nil;
    
    unsigned char* pOutBuf;
    int nOutWidth;
    int nOutHeight;
    int bufSize;
    
    //decode TPG
    
    
    int stream_len = 0;
    FILE* fp = fopen(InputFile, "rb");
    if (fp == NULL){
        return result;
    }
    struct stat fileInfo;
    if(stat(InputFile,&fileInfo))
    {
        return result;
    }
    stream_len = (int)fileInfo.st_size;
    unsigned char* pStreamBuf = (unsigned char*)malloc(stream_len);
    if(fread(pStreamBuf,1,stream_len,fp) !=stream_len)
    {
        free(pStreamBuf);
        pStreamBuf = NULL;
        return result;
    }
    fclose(fp);
    
    TPGFeatures features = {0};
    TPGStatusCode stats = TPGParseHeader(pStreamBuf, stream_len, &features);
    
    if (TPG_STATUS_OK != stats || features.image_mode != emMode_EncodeAlpha)
    {
        free(pStreamBuf);
        pStreamBuf = NULL;
        return result;
    }
    
    void *TPGDec = TPGDecCreate(pStreamBuf,stream_len);
    
    nOutWidth = features.width;
    nOutHeight = features.height;
    bufSize = nOutWidth*nOutHeight*4;
    pOutBuf = new unsigned char[bufSize];
    
    TPGOutFrame outFrame = {0};
    outFrame.dstWidth = nOutWidth;
    outFrame.dstHeight = nOutHeight;
    outFrame.pOutBuf = pOutBuf;
    outFrame.bufsize = bufSize;
    outFrame.fmt = enRawDataFormat::FORMAT_RGBA;
    
    

    TPGDecodeImage(TPGDec, pStreamBuf, stream_len, 0, &outFrame);
        

    
    if (pOutBuf){
               result = [UIImage convertBitmapRGBA8ToUIImage:pOutBuf withWidth:nOutWidth withHeight:nOutHeight];
        //result = [UIImage imageFromRGBABytes:pOutBuf imageSize:CGSizeMake(nOutWidth, nOutHeight)];
    }
    TPGDecDestroy(TPGDec);
    TPGDec = NULL;
    free(pStreamBuf);
    pStreamBuf = NULL;
    free(pOutBuf);
    pOutBuf = NULL;
    
    return result;
}

+ (UIImage *) convertBitmapRGBA8ToUIImage:(unsigned char *) buffer
                                withWidth:(int) width
                               withHeight:(int) height {
    
    
    size_t bytesPerRow = 4 * width;
    size_t componentsPerPix = 4;
    size_t bitsPerPixel         = 8;
    
    CGBitmapInfo bitmapInfo = kCGBitmapByteOrderDefault | kCGImageAlphaLast;
    CFDataRef data = CFDataCreate(kCFAllocatorDefault, buffer, width * height * componentsPerPix);
    CGDataProviderRef provider = CGDataProviderCreateWithCFData(data);
    CGColorSpaceRef colorSpace = CGColorSpaceCreateDeviceRGB();
    CGImageRef cgImage = CGImageCreate(width,
                                       height,
                                       bitsPerPixel,
                                       bitsPerPixel * componentsPerPix,
                                       bytesPerRow,
                                       colorSpace,
                                       bitmapInfo,
                                       provider,
                                       NULL,
                                       NO,
                                       kCGRenderingIntentDefault);
    CGColorSpaceRelease(colorSpace);
    UIImage *image = nil;
    if (nil != cgImage) {
        image = [UIImage imageWithCGImage:cgImage];
    }
    else {
        //QLog_Event(MODULE_IMPB_RICHMEDIA, "Error!! CGImageCreate failed!!!");
    }
    CGImageRelease(cgImage);
    CGDataProviderRelease(provider);
    CFRelease(data);
    return image;
}

+ (UIImage *)imageFromRGBABytes:(unsigned char *)imageBytes imageSize:(CGSize)imageSize {
    CGImageRef imageRef = [self imageRefFromRGBABytes:imageBytes imageSize:imageSize];
    UIImage *image = [UIImage imageWithCGImage:imageRef];
    CGImageRelease(imageRef);
    return image;
}

+ (CGImageRef)imageRefFromRGBABytes:(unsigned char *)imageBytes imageSize:(CGSize)imageSize {
    
    CGColorSpaceRef colorSpace = CGColorSpaceCreateDeviceRGB();
    CGContextRef context = CGBitmapContextCreate(imageBytes,
                                                 imageSize.width,
                                                 imageSize.height,
                                                 8,
                                                 imageSize.width * 4,
                                                 colorSpace,
                                                 kCGImageAlphaPremultipliedLast);
    CGImageRef imageRef = CGBitmapContextCreateImage(context);
    CGContextRelease(context);
    CGColorSpaceRelease(colorSpace);
    
    return imageRef;
}

@end
